@@ -15,18 +15,21 @@
 module Data.Acid.Local
     ( openLocalState
     , openLocalStateFrom
+    , prepareLocalState
+    , prepareLocalStateFrom
     , createArchive
     , createCheckpointAndClose
-    ) where 
+    ) where
 
 import Data.Acid.Log as Log
 import Data.Acid.Core
 import Data.Acid.Common
 import Data.Acid.Abstract
 
-import Control.Concurrent             ( newEmptyMVar, putMVar, takeMVar, MVar )
---import Control.Exception              ( evaluate )
+import Control.Concurrent             ( newEmptyMVar, putMVar, takeMVar, MVar, threadDelay )
+import Control.Exception              ( SomeException, handle )
 import Control.Monad.State            ( runState )
+import Control.Monad                  ( join )
 import Control.Applicative            ( (<$>), (<*>) )
 import Data.ByteString.Lazy           ( ByteString )
 --import qualified Data.ByteString.Lazy as Lazy ( length )
@@ -36,8 +39,11 @@ import Data.Serialize                 ( runPutLazy, runGetLazy )
 import Data.SafeCopy                  ( SafeCopy(..), safeGet, safePut
                                       , primitive, contain )
 import Data.Typeable                  ( Typeable, typeOf )
+import Data.IORef
 import System.FilePath                ( (</>) )
 
+import FileIO                         ( obtainPrefixLock, releasePrefixLock, PrefixLock )
+
 
 {-| State container offering full ACID (Atomicity, Consistency, Isolation and Durability)
     guarantees.
@@ -54,8 +60,10 @@ import System.FilePath                ( (</>) )
 -}
 data LocalState st
     = LocalState { localCore        :: Core st
+                 , localCopy        :: IORef st
                  , localEvents      :: FileLog (Tagged ByteString)
                  , localCheckpoints :: FileLog Checkpoint
+                 , localLock        :: PrefixLock
                  } deriving (Typeable)
 
 
@@ -79,7 +87,8 @@ scheduleLocalUpdate acidState event
            do let !(result, !st') = runState hotMethod st
               -- Schedule the log entry. Very important that it happens when 'localCore' is locked
               -- to ensure that events are logged in the same order that they are executed.
-              pushEntry (localEvents acidState) (methodTag event, encoded) $ putMVar mvar result
+              pushEntry (localEvents acidState) (methodTag event, encoded) $ do writeIORef (localCopy acidState) st'
+                                                                                putMVar mvar result
               return st'
          return mvar
     where hotMethod = lookupHotMethod (coreMethods (localCore acidState)) event
@@ -91,7 +100,8 @@ scheduleLocalColdUpdate acidState event
            do let !(result, !st') = runState coldMethod st
               -- Schedule the log entry. Very important that it happens when 'localCore' is locked
               -- to ensure that events are logged in the same order that they are executed.
-              pushEntry (localEvents acidState) event $ putMVar mvar result
+              pushEntry (localEvents acidState) event $ do writeIORef (localCopy acidState) st'
+                                                           putMVar mvar result
               return st'
          return mvar
     where coldMethod = lookupColdMethod (localCore acidState) event
@@ -99,27 +109,17 @@ scheduleLocalColdUpdate acidState event
 -- | Issue a Query event and wait for its result. Events may be issued in parallel.
 localQuery  :: QueryEvent event  => LocalState (EventState event) -> event -> IO (EventResult event)
 localQuery acidState event
-    = do mvar <- newEmptyMVar
-         withCoreState (localCore acidState) $ \st ->
-           do let (result, _st) = runState hotMethod st
-              -- Make sure that we do not return the result before the event log has
-              -- been flushed to disk.
-              pushAction (localEvents acidState) $
-                putMVar mvar result
-         takeMVar mvar
+    = do st <- readIORef (localCopy acidState)
+         let (result, _st) = runState hotMethod st
+         return result
     where hotMethod = lookupHotMethod (coreMethods (localCore acidState)) event
 
 -- Whoa, a buttload of refactoring is needed here. 2011-11-02
 localQueryCold  :: LocalState st -> Tagged ByteString -> IO ByteString
 localQueryCold acidState event
-    = do mvar <- newEmptyMVar
-         withCoreState (localCore acidState) $ \st ->
-           do let (result, _st) = runState coldMethod st
-              -- Make sure that we do not return the result before the event log has
-              -- been flushed to disk.
-              pushAction (localEvents acidState) $
-                putMVar mvar result
-         takeMVar mvar
+    = do st <- readIORef (localCopy acidState)
+         let (result, _st) = runState coldMethod st
+         return result
     where coldMethod = lookupColdMethod (localCore acidState) event
 
 -- | Take a snapshot of the state and save it to disk. Creating checkpoints
@@ -151,6 +151,7 @@ createCheckpointAndClose abstract_state
          takeMVar mvar
          closeFileLog (localEvents acidState)
          closeFileLog (localCheckpoints acidState)
+         releasePrefixLock (localLock acidState)
   where acidState = downcast abstract_state
 
 
@@ -172,8 +173,21 @@ openLocalState :: (Typeable st, IsAcidic st)
               => st                          -- ^ Initial state value. This value is only used if no checkpoint is
                                              --   found.
               -> IO (AcidState st)
-openLocalState initialState
-    = openLocalStateFrom ("state" </> show (typeOf initialState)) initialState
+openLocalState initialState =
+  openLocalStateFrom ("state" </> show (typeOf initialState)) initialState
+
+-- | Create an AcidState given an initial value.
+--
+--   This will create or resume a log found in the \"state\/[typeOf state]\/\" directory.
+--   The most recent checkpoint will be loaded immediately but the AcidState will not be opened
+--   until the returned function is executed.
+prepareLocalState :: (Typeable st, IsAcidic st)
+                  => st                          -- ^ Initial state value. This value is only used if no checkpoint is
+                                                 --   found.
+                  -> IO (IO (AcidState st))
+prepareLocalState initialState =
+  prepareLocalStateFrom ("state" </> show (typeOf initialState)) initialState
+
 
 -- | Create an AcidState given a log directory and an initial value.
 --
@@ -185,30 +199,74 @@ openLocalStateFrom :: (IsAcidic st)
                   -> st                  -- ^ Initial state value. This value is only used if no checkpoint is
                                          --   found.
                   -> IO (AcidState st)
-openLocalStateFrom directory initialState
-    = do core <- mkCore (eventsToMethods acidEvents) initialState
-         let eventsLogKey = LogKey { logDirectory = directory
-                                   , logPrefix = "events" }
-             checkpointsLogKey = LogKey { logDirectory = directory
-                                        , logPrefix = "checkpoints" }
-         mbLastCheckpoint <- Log.newestEntry checkpointsLogKey
-         n <- case mbLastCheckpoint of
-                Nothing
-                  -> return 0
-                Just (Checkpoint eventCutOff content)
-                  -> do modifyCoreState_ core (\_oldState -> case runGetLazy safeGet content of
-                                                               Left msg  -> checkpointRestoreError msg
-                                                               Right val -> return val)
-                        return eventCutOff
+openLocalStateFrom directory initialState =
+  join $ resumeLocalStateFrom directory initialState False
 
-         eventsLog <- openFileLog eventsLogKey
-         events <- readEntriesFrom eventsLog n
-         mapM_ (runColdMethod core) events
-         checkpointsLog <- openFileLog checkpointsLogKey
-         return $ toAcidState LocalState { localCore = core
-                                         , localEvents = eventsLog
-                                         , localCheckpoints = checkpointsLog
-                                         }
+-- | Create an AcidState given an initial value.
+--
+--   This will create or resume a log found in @directory@.
+--   The most recent checkpoint will be loaded immediately but the AcidState will not be opened
+--   until the returned function is executed.
+prepareLocalStateFrom :: (IsAcidic st)
+                  => FilePath            -- ^ Location of the checkpoint and transaction files.
+                  -> st                  -- ^ Initial state value. This value is only used if no checkpoint is
+                                         --   found.
+                  -> IO (IO (AcidState st))
+prepareLocalStateFrom directory initialState =
+  resumeLocalStateFrom directory initialState True
+
+
+
+resumeLocalStateFrom :: (IsAcidic st)
+                  => FilePath            -- ^ Location of the checkpoint and transaction files.
+                  -> st                  -- ^ Initial state value. This value is only used if no checkpoint is
+                                         --   found.
+                  -> Bool                -- ^ True => load checkpoint before acquiring the lock.
+                  -> IO (IO (AcidState st))
+resumeLocalStateFrom directory initialState delayLocking =
+  case delayLocking of
+    True -> do
+      (n, st) <- loadCheckpoint
+      return $ do
+        lock  <- obtainPrefixLock lockFile
+        replayEvents lock n st
+    False -> do
+      lock    <- obtainPrefixLock lockFile
+      (n, st) <- loadCheckpoint
+      return $ do
+        replayEvents lock n st
+  where
+    lockFile = directory </> "open"
+    eventsLogKey = LogKey { logDirectory = directory
+                          , logPrefix = "events" }
+    checkpointsLogKey = LogKey { logDirectory = directory
+                               , logPrefix = "checkpoints" }
+    loadCheckpoint = do
+      mbLastCheckpoint <- Log.newestEntry checkpointsLogKey
+      case mbLastCheckpoint of
+        Nothing ->
+          return (0, initialState)
+        Just (Checkpoint eventCutOff content) -> do
+          case runGetLazy safeGet content of
+            Left msg  -> checkpointRestoreError msg
+            Right val -> return (eventCutOff, val)
+    replayEvents lock n st = do
+      core <- mkCore (eventsToMethods acidEvents) st
+
+      eventsLog <- openFileLog eventsLogKey
+      events <- readEntriesFrom eventsLog n
+      mapM_ (runColdMethod core) events
+      ensureLeastEntryId eventsLog n
+      checkpointsLog <- openFileLog checkpointsLogKey
+      stateCopy <- newIORef undefined
+      withCoreState core (writeIORef stateCopy)
+
+      return $ toAcidState LocalState { localCore = core
+                                      , localCopy = stateCopy
+                                      , localEvents = eventsLog
+                                      , localCheckpoints = checkpointsLog
+                                      , localLock = lock
+                                      }
 
 checkpointRestoreError msg
     = error $ "Could not parse saved checkpoint due to the following error: " ++ msg
@@ -220,13 +278,14 @@ closeLocalState acidState
     = do closeCore (localCore acidState)
          closeFileLog (localEvents acidState)
          closeFileLog (localCheckpoints acidState)
+         releasePrefixLock (localLock acidState)
 
 
 -- | Move all log files that are no longer necessary for state restoration into the 'Archive'
 --   folder in the state directory. This folder can then be backed up or thrown out as you see fit.
 --   Reverting to a state before the last checkpoint will not be possible if the 'Archive' folder
 --   has been thrown out.
--- 
+--
 --   This method is idempotent and does not block the normal operation of the AcidState.
 createArchive :: AcidState st -> IO ()
 createArchive abstract_state
